import sys
if sys.version_info < (3, 4):
    raise ImportError("Transplant does not support Python < 3.4.")

from .transplant_master import Matlab, TransplantError, MatlabStruct, MatlabFunction, MatlabProxyObject
